@@ -9,7 +9,7 @@ import textwrap
 
 from collections import OrderedDict, namedtuple
 from itertools import takewhile
-from typing import Any, List, Optional, Union, Dict, Tuple
+from typing import Any, List, Optional, Union, Dict, Tuple, Sequence
 
 from bs4 import BeautifulSoup
 from bs4.element import Tag
@@ -29,7 +29,7 @@ TRAILING_COMMENT_START = ("/**<", "/*!<", "///<")
 
 #: Nodes which can be anonymous and still need to be documented. i.e.
 #: anonymous structures are usually part of another node. Anonymous enums are
-#: often used to use the enumerations but not force the type usage.
+#: often used to use the enumerators but not force the type usage.
 ALLOWED_ANONYMOUS = (cindex.CursorKind.ENUM_DECL,)
 
 # Must do this prior to calling into clang
@@ -43,6 +43,9 @@ class DocumentedObject:
     """
     A representation of a c object for documentation purposes.
 
+    Arguments:
+        node (:class:`~clang.cindex.Cursor`): The node representing this object.
+
     Attributes:
         type_ (str): The type of this item one of:
 
@@ -53,6 +56,9 @@ class DocumentedObject:
             - type: A typedef
             - struct: A structure
             - union: A Union
+            - enumerator: An enumerator constant
+            - macro: A macro
+
         doc (str): The default documentation of the object. This is usally
             the comment with leading '*' removed.
         name (str): The name of the object. For example functions this would
@@ -66,15 +72,15 @@ class DocumentedObject:
             the type as well as the name.
         _line_range (Tuple[int, int]): The line range of the C construct,
             this will include any leading or trailing comments that may be
-            part of the constructs documentation.
+            part of the construct's documentation.
     """
 
     type_ = "object"
 
-    def __init__(self) -> None:
+    def __init__(self, node: Cursor) -> None:
         self.doc = ""
         self.name = ""
-        self.node: Cursor = None
+        self.node = node
         self._children: Optional[OrderedDict] = None
         self._soup: Optional[BeautifulSoup] = None
         self._declaration: Optional[str] = None
@@ -85,7 +91,7 @@ class DocumentedObject:
         The lines in the source file that this object covers.
 
         This will include any leading or trailing comments that may be part
-        of the constructs documentation.
+        of the construct's documentation.
         """
         if self._line_range is None:
             node_extent = self.node.extent
@@ -146,11 +152,10 @@ class DocumentedObject:
         obj_dict["start_line"] = line_range[0]
         obj_dict["end_line"] = line_range[1]
 
-        if self.children:
-            obj_dict["children"] = []
+        obj_dict["children"] = []
 
-            for child in self.children.values():
-                obj_dict["children"].append(json.loads(str(child)))
+        for child in self.children.values():
+            obj_dict["children"].append(json.loads(str(child)))
 
         return json.dumps(obj_dict)
 
@@ -203,7 +208,12 @@ class DocumentedObject:
 
         root = self.soup.contents[0]
 
-        return root.declaration.text
+        # It seems with different versions of clang the newlines will at times
+        # be kept around from some declarations. This causes problems with
+        # sphinx as the signature should remain all in one line.
+        lines = root.declaration.text.splitlines()
+        declaration = " ".join(line.strip() for line in lines)
+        return declaration
 
     @property
     def soup(self) -> Optional[BeautifulSoup]:
@@ -279,23 +289,7 @@ class DocumentedMacro(DocumentedObject):
     A documented macro
     """
 
-    # Macros can be function like or they can be macro like so look up in the
-    # :attr:`type`.
-    _type = None
-
-    @property
-    def type(self) -> str:
-        """
-        Type of this object
-        :return: The type
-        """
-        if self._type is None:
-            if self.node.is_macro_function_like():
-                self._type = "function"
-            else:
-                self._type = "macro"
-
-        return self._type
+    type_ = "macro"
 
     def format_args(self, **kwargs: Any) -> str:
         """
@@ -333,7 +327,7 @@ class DocumentedMacro(DocumentedObject):
         Creates the full declaration of the macro. For function like macros
         this will include the parenthesised arguments.
         """
-        if self.type == "macro":
+        if not self.node.is_macro_function_like():
             return f"{self.name}"
 
         # We know this must be a function like macro, which means the first 2
@@ -351,7 +345,15 @@ class DocumentedMacro(DocumentedObject):
             i.spelling for i in ident_iter if i.kind == cindex.TokenKind.IDENTIFIER
         ]
 
-        return "{}({})".format(self.name, ", ".join(tokens))
+        return f"{self.name}({', '.join(tokens)})"
+
+
+class DocumentedEnumerator(DocumentedMacro):
+    """
+    An enumerator, the constant values in an enum
+    """
+
+    type_ = "enumerator"
 
 
 class DocumentedMember(DocumentedObject):
@@ -375,7 +377,13 @@ class DocumentedMember(DocumentedObject):
         The parsed declaration of `bar` would be `int bar`.
         """
         type_ = self.node.type.spelling
-        return f"{type_} {self.name}"
+
+        # Libclang will return the type as `float [20]` when looking at
+        # `float foo[20]`.  We could look at the kind `TypeKind.CONSTANTARRAY`
+        # but partitioning on the "[" just seems more straight forward.
+        type_, *(array) = type_.partition("[")
+        array_contents = "".join(array)
+        return f"{type_} {self.name} {array_contents}"
 
     def is_public(self) -> bool:
         """
@@ -408,6 +416,10 @@ class DocumentedFunction(DocumentedObject):
         body = self.get_paragraph(root.find("abstract", recursive=False))
         body += self.get_paragraph(root.find("discussion", recursive=False))
 
+        # Single newlines are tread as the same paragraph in restructured text,
+        # providing 2 results in separate paragraphs.
+        body = body.replace("\n", "\n\n")
+
         for param in root.find_all("parameter"):
             name = param.find("name").text
             param_doc = self.get_paragraph(param.discussion)
@@ -451,16 +463,53 @@ class DocumentedFunction(DocumentedObject):
         name, _ = decl.split("(", 1)
         return name
 
+    def _get_arguments(self) -> str:
+        """
+        Get the arguments for this instance.
+
+        Returns:
+            str: The arguments for use in the function signature.
+        """
+        func = self.node
+
+        # Early logic used to iterate over, `func.get_arguments()`, however when there
+        # is an unknown type clang will sometimes fail to provide tokens for that
+        # argument. For example in "unknown_type foo[]" the brackets will cause clang
+        # to return back no tokens for the argument.
+        start = func.location
+        end = func.extent.end
+        if func.is_definition():
+            # When a function is a definition the last child is the compound statement
+            # so we need to move prior to the compound statement
+            children = list(func.get_children())
+            body_start = children[-1].extent.start.offset
+            end = cindex.SourceLocation.from_offset(func.tu, start.file, body_start - 1)
+
+        extent = cindex.SourceRange.from_locations(start, end)
+        non_comment_tokens = (
+            t
+            for t in cindex.TokenGroup.get_tokens(func.tu, extent=extent)
+            if t.kind != cindex.TokenKind.COMMENT
+        )
+
+        # Even though this will place spaces around all the tokens, the sphinx C domain
+        # will provide some formatting to make it look nicer in the final output.
+        full_signature = " ".join(t.spelling for t in non_comment_tokens)
+
+        _, _, arguments = full_signature.partition("(")
+        arguments = arguments.rstrip(")")
+        arguments = arguments.strip()
+
+        return arguments
+
     def get_parsed_declaration(self) -> str:
         """
         Creates the parenthesis version of the function signature.  i.e. this
         will be the `(int hello, int what)` portion of the header.
         """
-        func = self.node
-        args = []
-        for arg in func.get_arguments():
-            args.append(" ".join(t.spelling for t in arg.get_tokens()))
+        args = self._get_arguments()
 
+        func = self.node
         tu = func.tu
 
         # For functions the extent encompasses the return value, and the
@@ -475,7 +524,7 @@ class DocumentedFunction(DocumentedObject):
             t.spelling for t in cindex.TokenGroup.get_tokens(tu, extent=extent)
         )
 
-        return "{} {}({})".format(return_type, func.spelling, ", ".join(args))
+        return f"{return_type} {func.spelling}({args})"
 
     def is_public(self) -> bool:
         """
@@ -504,6 +553,27 @@ class DocumentedType(DocumentedObject):
         For things like functions and others this will include the return type.
         """
         parent_type = self.node.underlying_typedef_type.spelling
+
+        # Function prototypes need to be handled different. When clang can't
+        # successfully parse the file it falls back to naming the return type
+        # as the display name.
+        # Unfortunately some versions of clang behave a little differently, some
+        # will return a `POINTER` while others will return `FUNCITONNOPROTO`. The
+        # `POINTER`s are easy to derive the real type from, but the test
+        # environment doesn't use that version of clang.
+        type_ = self.node.underlying_typedef_type
+        if type_.kind == cindex.TypeKind.POINTER:  # pragma: no cover
+            type_ = type_.get_pointee()
+
+        if type_.kind in (
+            cindex.TypeKind.FUNCTIONPROTO,
+            cindex.TypeKind.FUNCTIONNOPROTO,
+        ):
+            ret_value, paren, signature = parent_type.partition(")")
+            signature = "".join((ret_value, self.name, paren, signature))
+
+            return f"typedef {signature}"
+
         return f"typedef {parent_type} {self.name}"
 
 
@@ -518,16 +588,16 @@ class DocumentedStructure(DocumentedObject):
     def soup(self) -> None:
         """
         Since structures like objects use the "Members:" and
-        "Enumerations:" sections do *not* use the clang xml comments as they
-        don't preseve newlines, so the sections get lost.
+        "Enumerators:" sections do *not* use the clang xml comments as they
+        don't preserve newlines, so the sections get lost.
         """
         return None
 
     def get_parsed_declaration(self) -> str:
         """
-        Structures, and similar, are just the type and the name.
+        Structures, and similar, are just name.
         """
-        return f"{self.type} {self.name}"
+        return f"{self.name}"
 
     @property
     def children(self) -> dict:
@@ -556,7 +626,7 @@ class DocumentedUnion(DocumentedStructure):
 
 class DocumentedEnum(DocumentedStructure):
     """
-    Class for Enumerations. Same as structures with a different :attr:`type`.
+    Class for Enums. Same as structures with a different :attr:`type`.
     """
 
     type_ = "enum"
@@ -595,6 +665,54 @@ class DocumentedVariable(DocumentedObject):
 
         return True
 
+    def get_parsed_declaration(self) -> str:
+        """
+        Get the declaration as parsed from the :attr:`node`.
+        """
+        # Libclang will return the type as `float [20]` when looking at
+        # `float foo[20]`.  We could look at the kind `TypeKind.CONSTANTARRAY`
+        # but partitioning on the "[" just seems more straight forward.
+        type_ = self.node.type.spelling
+        type_, *(array) = type_.partition("[")
+        array_contents = "".join(array)
+
+        real_type = self._find_declaration_type()
+        type_ = type_.replace("int", real_type)
+
+        return f"{type_} {self.name} {array_contents}"
+
+    def _find_declaration_type(self) -> str:
+        """
+        Makes an attempt to try and find the identifier, and storage class,
+        representing the variable type.
+
+        Returns:
+            str: The type of the variable.  If this can't be derived falls back to
+                `int`.
+        """
+        type_ = "int"
+        tokens = list(
+            filter(
+                lambda t: t.kind == cindex.TokenKind.IDENTIFIER, self.node.get_tokens()
+            )
+        )
+        try:
+            type_ = tokens[-2].spelling
+        except IndexError:
+            # For array variables with unknown types libclang fails to provide the
+            # tokens.
+            pass
+
+        # clang doesn't provide the storage class in the type name, so we'll add it here
+        storage_keyword = ""
+        storage_class = self.node.storage_class
+        if storage_class == cindex.StorageClass.STATIC:
+            storage_keyword = "static"
+        if storage_class == cindex.StorageClass.EXTERN:
+            storage_keyword = "extern"
+
+        return f"{storage_keyword} {type_}"
+
 
 CURSORKIND_TO_OBJECT_CLASS = {
     cindex.CursorKind.TRANSLATION_UNIT: DocumentedFile,
@@ -604,7 +722,7 @@ CURSORKIND_TO_OBJECT_CLASS = {
     cindex.CursorKind.ENUM_DECL: DocumentedEnum,
     cindex.CursorKind.FIELD_DECL: DocumentedMember,
     cindex.CursorKind.MACRO_DEFINITION: DocumentedMacro,
-    cindex.CursorKind.ENUM_CONSTANT_DECL: DocumentedMacro,
+    cindex.CursorKind.ENUM_CONSTANT_DECL: DocumentedEnumerator,
     cindex.CursorKind.VAR_DECL: DocumentedVariable,
     cindex.CursorKind.TYPEDEF_DECL: DocumentedType,
 }
@@ -622,7 +740,9 @@ def object_from_cursor(cursor: Cursor) -> Optional[DocumentedObject]:
         # type spelling of:
         # "<construct> (anonymous at # <path_to_c_file>:<lineno>)"
         # Typedef's should be handled by the get_nested_node() function
-        if cursor.kind in ALLOWED_ANONYMOUS and "anonymous at" in cursor.type.spelling:
+        if cursor.kind in ALLOWED_ANONYMOUS and any(
+            anon in cursor.type.spelling for anon in ("anonymous at", "unnamed at")
+        ):
             filename = os.path.basename(cursor.location.file.name)
             # remove the extension from the filename since the '.' is not a
             # valid c identifier. splitext will remove the trailing most
@@ -636,14 +756,13 @@ def object_from_cursor(cursor: Cursor) -> Optional[DocumentedObject]:
 
     nested_cursor = get_nested_node(cursor)
     class_ = CURSORKIND_TO_OBJECT_CLASS.get(nested_cursor.kind, DocumentedObject)
-    doc = class_()
+    doc = class_(nested_cursor)
 
     doc.name = name
     psuedo_comment = PsuedoToken(
         nested_cursor.raw_comment, nested_cursor.comment_extent
     )
     doc.doc = parse_comment(psuedo_comment)
-    doc.node = nested_cursor
 
     return doc
 
@@ -661,6 +780,7 @@ def get_nested_node(cursor: Cursor) -> Cursor:
             underlying_node = next(cursor.get_children())
             if underlying_node.kind in (
                 cindex.CursorKind.STRUCT_DECL,
+                cindex.CursorKind.UNION_DECL,
                 cindex.CursorKind.ENUM_DECL,
             ):
                 return underlying_node
@@ -703,27 +823,76 @@ def get_file_comment(cursor: Cursor, child: Optional[Cursor]) -> str:
     return ""
 
 
-def load(filename: str, contents: str) -> DocumentedObject:
+def get_compilation_args(
+    filename: str, compilation_database: Optional[str] = None
+) -> List[str]:
+    """
+    Get the compilation args for `filename` for the compilation database found in
+    `compilation_db_dir`
+
+    Args:
+        filename (str): The file to get the compilation arguments for.
+        compilation_database (str): The compilation database.
+
+    Returns:
+        list[str]: The compilation arguments.
+    """
+    if not compilation_database:
+        return []
+    directory = os.path.dirname(compilation_database)
+    comp_db = cindex.CompilationDatabase.fromDirectory(directory)
+    commands = comp_db.getCompileCommands(filename)
+
+    if not commands:
+        return []
+
+    # For now only handling the first file instance seen in the database.
+    # First argument is compiler path, last is the file to compile
+    args = list(commands[0].arguments)[1:-1]
+
+    # Since things like includes and defines could be relative we force the working
+    # directory.
+    working_dir = commands[0].directory
+    args.append(f"-working-directory={working_dir}")
+
+    return args
+
+
+def load(
+    filename: str,
+    contents: str,
+    compilation_database: Optional[str] = None,
+    compilation_args: Optional[Sequence[str]] = None,
+) -> DocumentedObject:
     """
     Load a C file into a tree of :class:`DocumentedObject`\'s
 
     Args:
         filename (str): The c file to load into a documented item
         contents (str): The contents of `filename`
+        compilation_database (str): The compilation database.
+        compilation_args (str): Compilation arguments.  Will be applied *after*
+            compilation database.
 
     Returns:
         :class:`DocumentedObject`: The documented version of `filename`.
 
     """
+    args = get_compilation_args(filename, compilation_database)
+    if compilation_args:
+        args += compilation_args
 
     tu = cindex.TranslationUnit.from_source(
         filename,
-        unsaved_files=((filename, contents),),
+        args=args,
+        unsaved_files=[
+            (filename, contents),
+        ],
         options=cindex.TranslationUnit.PARSE_DETAILED_PROCESSING_RECORD,
     )
     cursor = tu.cursor
 
-    root_document = DocumentedFile()
+    root_document = DocumentedFile(cursor)
 
     # Some nodes show up from header includes as well as compiler defines, so
     # skip those. Macro instantiations are the locations where macros are
@@ -740,7 +909,7 @@ def load(filename: str, contents: str) -> DocumentedObject:
     ]
 
     # Macro definitions always come first in the child list, but that may not
-    # be their location in the file, so sort all of the nodes by location
+    # be their location in the file, so sort all the nodes by location
     sorted_nodes = sorted(child_nodes, key=lambda x: x.extent.start.offset)
 
     comment_nodes(cursor, sorted_nodes)
